import functools

import numpy as np


def convert(source):
  if callable(source):
    return source
  elif hasattr(source, '__getitem__'):
    return source.__getitem__
  else:
    raise TypeError(source)


class Transform:

  def __init__(self, source, fn, seed=0):
    self.source = convert(source)
    self.fn = fn
    self.seed = seed

  def __call__(self, step):
    datapoint = self.source(step)
    datapoint = self.fn(datapoint, seed=[self.seed, step])
    return datapoint


class Sample:

  def __init__(self, source, length=None, seed=0):
    length = length or len(source)
    source = convert(source)
    self.source = source
    self.length = length
    self.seed = seed

  def __call__(self, step):
    rng = np.random.default_rng(seed=[self.seed, step])
    index = rng.choice(self.length)
    return self.source(index)


class Epochs:

  def __init__(self, source, length=None, shuffle=True, seed=0):
    length = length or len(source)
    source = convert(source)
    self.source = source
    self.length = length
    self.shuffle = shuffle
    self.seed = 0

  def __call__(self, step):
    epoch = step // self.length
    index = step % self.length
    if self.shuffle:
      index = self._permutation(epoch)[index]
    return self.source(index)

  @functools.lru_cache(maxsize=1)
  def _permutation(self, epoch):
    rng = np.random.default_rng(seed=[self.seed, epoch])
    return rng.permutation(np.arange(self.length)).tolist()


class Truncate:

  def __init__(self, source, limit, length=None):
    length = length or len(source)
    self.source = convert(source)
    self.limit = min(length, limit)

  def __call__(self, step):
    return self.source(step % self.limit)


class Interleave:

  def __init__(self, sources):
    self.sources = [convert(source) for source in sources]

  def __call__(self, step):
    outer = step % len(self.sources)
    inner = step // len(self.sources)
    return self.sources[outer](inner)


class Mix:

  def __init__(self, sources, weights, seed=0):
    self.sources = [convert(source) for source in sources]
    weights = np.asarray(weights, np.float32)
    weights /= weights.sum()
    self.weights = weights
    self.seed = seed

  def __call__(self, step):
    rng = np.random.default_rng(seed=[self.seed, step])
    choice = rng.choice(len(self.sources), p=self.weights)
    return self.sources[choice](step)



class Closing:

  def __init__(self):
    self.closed = False

  def __enter__(self):
    # assert not self.closed
    return self

  def __exit__(self, *e):
    try:
      self.close()
    except Exception:
      if not e[0]:
        raise
    self.closed = True



import errno
import io
import os
import pathlib
import pickle
import struct
from multiprocessing import shared_memory

from . import utils


limst = struct.Struct('<Q')


class BagWriter(utils.Closing):

  def __init__(self, filepath, version=2):
    assert version in (1, 2), version
    super().__init__()
    if isinstance(filepath, str):
      filepath = pathlib.Path(filepath)
    assert not filepath.exists(), filepath
    file = filepath.open('wb')
    assert isinstance(file, io.BufferedIOBase), type(file)
    assert file.writable(), file
    # Version 1 stores the starts and ends of all blobs, resulting in N+1
    # entries in the limits table, where the first entry is zero.
    # Version 2 stores only the ends of all blobs, resulting in N entries in
    # the limits table.
    self.version = version
    self.file = file
    self.offset = 0
    self.length = 0
    self.limits = []
    self.towrite = []

  @property
  def size(self):
    return self.offset + 8 * self.length + (8 if self.version == 1 else 0)

  def __len__(self):
    return self.length

  def append(self, record, flush=True):
    assert len(record), 'zero byte record'
    assert not self.closed
    assert self.length < 2 ** 32 - 1, self.length
    assert isinstance(record, bytes), type(record)
    index = self.length
    self.length += 1
    if self.version == 1:
      self.limits.append(self.offset)
      self.offset += len(record)
    if self.version == 2:
      self.offset += len(record)
      self.limits.append(self.offset)
    self.towrite.append(record)
    flush and self.flush()
    return index

  def flush(self):
    assert not self.closed
    if not self.towrite:
      return
    if len(self.towrite) == 1:
      self.file.write(self.towrite[0])
    else:
      self.file.write(b''.join(self.towrite))
    self.towrite.clear()

  def close(self):
    assert not self.closed
    if self.version == 1:
      self.limits.append(self.offset)
    self.towrite.extend(limst.pack(x) for x in self.limits)
    self.flush()
    self.file.close()


class BagReader(utils.Closing):

  def __init__(self, source, cache_index=True, cache_data=False):
    super().__init__()
    if isinstance(source, str):
      source = pathlib.Path(source)
    if cache_data and hasattr(source, 'open'):
      cache_index = False
      with source.open('rb') as f:
        source = SharedBuffer(f.read())
    if not hasattr(source, 'open'):
      cache_index, cache_data = False, False
      source = SharedBuffer(source)
    self.source = source
    self.fp = None
    assert self.file.readable(), self.file
    try:
      self.file.seek(-8, os.SEEK_END)
      empty = False
    except OSError as e:
      if e.errno != errno.EINVAL:
        raise
      empty = True
    if empty:
      self.filesize = 0
      self.indexstart = 0
      self.length = 0
    else:
      self.filesize = self.file.tell() + 8
      self.indexstart = limst.unpack(self.file.read(8))[0]
      # If the index table starts with a zero, we are reading a version 1 file
      # and we can skip the initial entry.
      self.file.seek(self.indexstart, os.SEEK_SET)
      if limst.unpack(self.file.read(8))[0] == 0:
        self.indexstart += 8
      self.length = (self.filesize - self.indexstart) // 8
    if cache_index and self.length:
      self.file.seek(self.indexstart, os.SEEK_SET)
      limits = self.file.read(self.filesize - self.indexstart)
      self.limits = SharedBuffer(limits)
    else:
      self.limits = None

  @property
  def file(self):
    if not self.fp:
      self.fp = self.source.open('rb')
    return self.fp

  def __getstate__(self):
    return {
        'source': self.source,
        'filesize': self.filesize,
        'indexstart': self.indexstart,
        'length': self.length,
        'limits': self.limits,
    }

  def __setstate__(self, d):
    self.source = d['source']
    self.filesize = d['filesize']
    self.indexstart = d['indexstart']
    self.length = d['length']
    self.limits = d['limits']
    self.closed = False
    self.fp = None

  @property
  def size(self):
    return self.filesize

  def __len__(self):
    return self.length

  def __getitem__(self, index):
    assert not self.closed
    assert isinstance(index, (int, slice, range)), (index, type(index))
    if isinstance(index, int):
      lhs, rhs = self._getlims(index, index + 1)
      self.file.seek(lhs, os.SEEK_SET)
      return self.file.read(rhs - lhs)
    else:
      assert 0 <= index.start <= index.stop and index.step in (None, 1), index
      index = range(index.start, min(index.stop, self.length))
      if index.start == index.stop:
        return []
      limits = self._getlims(index.start, index.stop)
      lhs, rhs = limits[0], limits[-1]
      self.file.seek(lhs, os.SEEK_SET)
      buffer = self.file.read(rhs - lhs)
      records = [
          buffer[i - lhs: j - lhs]
          for i, j in zip(limits[:-1], limits[1:])]
      return records

  def copy(self):
    return pickle.loads(pickle.dumps(self))

  def close(self):
    assert not self.closed
    if self.fp:
      self.fp.close()
    if isinstance(self.limits, SharedBuffer):
      self.limits.close()
    if isinstance(self.source, SharedBuffer):
      self.source.close()

  def _getlims(self, start, stop):
    assert start < stop, (start, stop)
    lhs = 8 * max(0, start - 1)
    rhs = 8 * stop
    if self.limits:
      buffer = self.limits[lhs: rhs]
    else:
      self.file.seek(self.indexstart + lhs, os.SEEK_SET)
      buffer = self.file.read(rhs - lhs)
    limits = [
        limst.unpack(buffer[i: i + 8])[0]
        for i in range(0, len(buffer), 8)]
    return limits if start else (0, *limits)


class SharedBuffer:

  ENABLE = True

  def __init__(self, content):
    if self.ENABLE:
      self.shm = shared_memory.SharedMemory(create=True, size=len(content))
      self.shm.buf[:] = memoryview(content)
      self.buf = self.shm.buf
    else:
      self.buf = bytes(content)

  def __getitem__(self, index):
    return self.buf[index]

  def __getstate__(self):
    if self.ENABLE:
      return self.shm.name
    else:
      return self.buf

  def __setstate__(self, value):
    if self.ENABLE:
      self.shm = shared_memory.SharedMemory(name=value)
      self.buf = self.shm.buf
    else:
      self.buf = value

  def open(self, mode='rb'):
    assert mode in ('rb', 'wb'), mode
    return io.BytesIO(self.buf)

  def close(self):
    self.buf = None
    try:
      self.shm.unlink()
    except FileNotFoundError:
      pass



import multiprocessing
import pathlib
import queue

import granular
import numpy as np
import pytest


class TestLoader:

  @pytest.mark.parametrize('batch', (1, 3, 6, 12))
  def test_ordered(self, tmpdir, batch):
    directory = pathlib.Path(tmpdir) / 'dataset'
    spec = {'foo': 'int', 'bar': 'array'}
    datapoints = [{'foo': i, 'bar': np.ones((3, 3))} for i in range(10)]
    with granular.DatasetWriter(directory, spec, granular.encoders) as writer:
      [writer.append(x) for x in datapoints]
    reader = granular.DatasetReader(directory, granular.decoders)
    source = granular.sources.Epochs(reader, len(reader), shuffle=False)
    dataset = iter(granular.Loader(source, batch, workers=4))
    for i in range(0, 2 * len(datapoints), batch):
      data = next(dataset)
      assert set(data.keys()) == {'foo', 'bar'}
      assert data['foo'].shape == (batch,)
      assert data['bar'].shape == (batch, 3, 3)
      assert (data['foo'] == (
          np.arange(i, i + batch) % len(datapoints))).all()
      assert (data['bar'] == np.ones((batch, 3, 3))).all()
    dataset.close()
    reader.close()

  @pytest.mark.parametrize('recycle_after', (False, 1, 2))
  def test_recycle(self, tmpdir, recycle_after, batch=3):
    directory = pathlib.Path(tmpdir) / 'dataset'
    spec = {'foo': 'int', 'bar': 'array'}
    datapoints = [{'foo': i, 'bar': np.ones((3, 3))} for i in range(10)]
    with granular.DatasetWriter(directory, spec, granular.encoders) as writer:
      [writer.append(x) for x in datapoints]
    reader = granular.DatasetReader(directory, granular.decoders)
    source = granular.sources.Epochs(reader, len(reader), shuffle=False)
    dataset = iter(granular.Loader(
        source, batch, workers=4, recycle_after=recycle_after))
    for i in range(0, 6 * len(datapoints), batch):
      data = next(dataset)
      assert set(data.keys()) == {'foo', 'bar'}
      assert data['foo'].shape == (batch,)
      assert data['bar'].shape == (batch, 3, 3)
      assert (data['foo'] == (
          np.arange(i, i + batch) % len(datapoints))).all()
      assert (data['bar'] == np.ones((batch, 3, 3))).all()
    dataset.close()
    reader.close()

  @pytest.mark.parametrize('batch', (1, 3))
  def test_shuffled(self, tmpdir, batch):
    directory = pathlib.Path(tmpdir) / 'dataset'
    spec = {'foo': 'int', 'bar': 'array'}
    datapoints = [{'foo': i, 'bar': np.ones((3, 3))} for i in range(10)]
    with granular.DatasetWriter(directory, spec, granular.encoders) as writer:
      [writer.append(x) for x in datapoints]
    reader = granular.DatasetReader(directory, granular.decoders)
    source = granular.sources.Epochs(reader, len(reader), shuffle=True)
    dataset = iter(granular.Loader(source, batch, workers=4))
    seen = set()
    for i in range(0, len(datapoints), batch):
      data = next(dataset)
      assert (data['bar'] == np.ones((batch, 3, 3))).all()
      if i + batch <= len(datapoints):
        assert all([x not in seen for x in data['foo']])
        seen.update(data['foo'])
      else:
        seen.clear()
    dataset.close()
    reader.close()

  @pytest.mark.parametrize('kwargs', (
      dict(cache_index=False, cache_keys=[], parallel=False),
      dict(cache_index=True, cache_keys=['refs'], parallel=True),
  ))
  def test_reader_options(self, tmpdir, kwargs, batch=3):
    directory = pathlib.Path(tmpdir) / 'dataset'
    spec = {'foo': 'int', 'bar': 'array'}
    datapoints = [{'foo': i, 'bar': np.ones((3, 3))} for i in range(10)]
    with granular.DatasetWriter(directory, spec, granular.encoders) as writer:
      [writer.append(x) for x in datapoints]
    reader = granular.DatasetReader(directory, granular.decoders, **kwargs)
    source = granular.sources.Epochs(reader, len(reader), shuffle=True)
    dataset = iter(granular.Loader(source, batch, workers=4))
    seen = set()
    for i in range(0, len(datapoints), batch):
      data = next(dataset)
      assert (data['bar'] == np.ones((batch, 3, 3))).all()
      if i + batch <= len(datapoints):
        assert all([x not in seen for x in data['foo']])
        seen.update(data['foo'])
      else:
        seen.clear()
    dataset.close()
    reader.close()

  def test_shared_array_pool(self):
    import granular.loader
    mp = multiprocessing.get_context('spawn')
    array = granular.loader.SharedArray((10, 8), np.int32)
    with mp.Pool(4) as pool:
      list(pool.map(fill1, ((array, i) for i in reversed(range(10)))))
    result = array.result()
    assert result.shape == (10, 8)
    assert (result == np.arange(10)[:, None]).all()

  def test_shared_array_queue(self):
    import granular.loader
    mp = multiprocessing.get_context('spawn')
    array = granular.loader.SharedArray((10, 8), np.int32)
    stop = mp.Event()
    iqueue = mp.Queue()
    oqueue = mp.Queue()
    args = (stop, iqueue, oqueue)
    workers = [mp.Process(target=fill2, args=args) for _ in range(4)]
    [x.start() for x in workers]
    for i in reversed(range(10)):
      iqueue.put((array, i))
    for i in reversed(range(10)):
      oqueue.get()
    stop.set()
    [x.join() for x in workers]
    result = array.result()
    assert result.shape == (10, 8)
    assert (result == np.arange(10)[:, None]).all()

  @pytest.mark.parametrize('running', (True, False))
  def test_checkpoint(self, tmpdir, running, batch=2):
    directory = pathlib.Path(tmpdir) / 'dataset'
    spec = {'foo': 'int'}
    datapoints = [{'foo': i} for i in range(10)]
    with granular.DatasetWriter(directory, spec, granular.encoders) as writer:
      [writer.append(x) for x in datapoints]
    reader = granular.DatasetReader(directory, granular.decoders)
    source = granular.sources.Epochs(reader, len(reader), shuffle=True)
    dataset = iter(granular.Loader(source, batch, workers=4))
    source = granular.DatasetReader(directory, granular.decoders, **kwargs)
    for i in range(0, 2 * batch, batch):
      assert np.all(next(dataset)['foo'] == np.arange(i, i + batch) % 10)
    assert dataset.consumed == 2 * batch
    state = dataset.save()
    dataset.close()
    reader.close()
    del dataset
    del source
    reader = granular.DatasetReader(directory, granular.decoders)
    source = granular.sources.Epochs(reader, len(reader), shuffle=True)
    dataset = granular.Loader(source, batch, workers=4)
    if running:
      dataset = iter(dataset)
      dataset.load(state)
    else:
      dataset.load(state)
      dataset = iter(dataset)
    assert dataset.consumed == 2 * batch
    for i in range(2 * batch, 6 * batch, batch):
      assert np.all(next(dataset)['foo'] == np.arange(i, i + batch) % 10)
    assert dataset.consumed == 6 * batch
    dataset.close()
    reader.close()

  @pytest.mark.parametrize('length', (2, 10))
  @pytest.mark.parametrize('num_shards', (1, 2, 3))
  @pytest.mark.parametrize('local_batch', (1, 3, 6))
  def test_sharded(self, length, num_shards, local_batch):
    global_batch = local_batch * num_shards
    source = [{'foo': i} for i in range(length)]
    source = granular.sources.Epochs(source, length=1, shuffle=False)
    loaders = []
    for shard_id in range(num_shards):
      loaders.append(granular.Loader(
          source, local_batch, workers=2,
          shard_id=shard_id, num_shards=num_shards))
    step = 0
    for _, batches in zip(range(12), zip(*loaders)):
      actual = np.concatenate([x['foo'] for x in batches])
      should = np.arange(step, step + global_batch) % length
      assert actual.shape == should.shape and np.all(actual == should)
      step += global_batch
    [x.close() for x in loaders]

  @pytest.mark.parametrize('num_shards', (2, 3))
  @pytest.mark.parametrize('local_batch', (1, 2, 12))
  def test_sharded_checkpoint(self, num_shards, local_batch, length=10):
    global_batch = local_batch * num_shards
    source = [{'foo': i} for i in range(length)]
    source = granular.sources.Epochs(source, length=1, shuffle=False)
    step = 0
    make = lambda i: granular.Loader(
        source, local_batch, workers=2,
        shard_id=i, num_shards=num_shards)
    loaders = [make(i) for i in range(num_shards)]
    for _, batches in zip(range(12), zip(*loaders)):
      actual = np.concatenate([x['foo'] for x in batches])
      should = np.arange(step, step + global_batch) % length
      assert actual.shape == should.shape and np.all(actual == should)
      step += global_batch
    states = [x.save() for x in loaders]
    [x.close() for x in loaders]
    loaders = [make(i) for i in range(num_shards)]
    [x.load(s) for x, s in zip(loaders, states)]
    for _, batches in zip(range(12), zip(*loaders)):
      actual = np.concatenate([x['foo'] for x in batches])
      should = np.arange(step, step + global_batch) % length
      assert actual.shape == should.shape and np.all(actual == should)
      step += global_batch
    [x.close() for x in loaders]


def fill1(args):
  target, i = args
  target.array[i] = i


def fill2(stop, iqueue, oqueue):
  while not stop.is_set():
    try:
      message = iqueue.get(timeout=0.1)
    except queue.Empty:
      continue
    target, i = message
    target.array[i] = i
    oqueue.put(i)



import granular


class TestSources:

  def test_transform(self):
    data = list(range(10))
    source = granular.sources.Transform(data, (lambda x, seed: 2 * x))
    output = [source(step) for step in range(10)]
    assert output == [2 * x for x in data]

  def test_sample(self):
    data = list(range(10))
    source = granular.sources.Sample(data)
    output = [source(step) for step in range(10)]
    assert output != data
    assert sorted(output) != data
    assert set(output) <= set(data)

  def test_truncate(self):
    data = list(range(10))
    source = granular.sources.Truncate(data, 5)
    output = [source(step) for step in range(10)]
    assert output == [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]

  def test_epochs_ordered(self):
    data = list(range(10))
    source = granular.sources.Epochs(data, shuffle=False)
    output = [source(step) for step in range(20)]
    assert output[:10] == data
    assert output[10:] == data

  def test_epochs_shuffle(self):
    data = list(range(10))
    source = granular.sources.Epochs(data, shuffle=True)
    output = [source(step) for step in range(20)]
    assert output[:10] != data
    assert output[10:] != data
    assert sorted(output[:10]) == data
    assert sorted(output[10:]) == data

  def test_interleave(self):
    data1 = list(range(10))
    data2 = list('abcdefghij')
    source = granular.sources.Interleave([data1, data2])
    output = [source(step) for step in range(20)]
    assert output[::2] == data1
    assert output[1::2] == data2

  def test_mix(self, thres=0.2):
    data1 = list(range(10))
    data2 = list('abcdefghij')
    source1 = granular.sources.Sample(data1)
    source2 = granular.sources.Sample(data2)
    source = granular.sources.Mix([source1, source2], [0.2, 0.8])
    output = [source(step) for step in range(1000)]
    output1 = [x for x in output if x in data1]
    output2 = [x for x in output if x in data2]
    assert set(output) == set(data1 + data2)
    frac1 = len(output1) / len(output)
    frac2 = len(output2) / len(output)
    assert (1 - thres) * 0.2 <= frac1 <= (1 + thres) * 0.2
    assert (1 - thres) * 0.8 <= frac2 <= (1 + thres) * 0.8
